{$APPTYPE CONSOLE}

{-----------------------------------------------------------------------------}
{ NOTE: this code tested with FPC v3.0.4 and Delphi 10 (Seattle)              }
{-----------------------------------------------------------------------------}

program RangeChecks;

const
  SPACE = ' ';

var
  AnArray : array[0..0] of ansichar;

  b       : integer;
  AnInt   : integer;         // 32bit integer

  AnInt64 : int64;           // 64bit integer

begin
  AnArray[0]  := SPACE;      // no problem here

  {$ifdef FPC}
    // -------------------------------------------------------------------------
    // case 1.
    // Delphi emits an error for this expression (as it should)
    // FPC emits a warning but, at least, generates CORRECT code.

    AnArray[5]  := SPACE;    // Delphi won't compile this (which is correct)


    // -------------------------------------------------------------------------
    // case 2.
    // same as above for this case

    AnArray[-5] := SPACE;    // nor this but, FPC does. At least, in this case
                             // it generates CORRECT code for it.
  {$endif}

  // if FPC did range checking the way it should be done, the above statements
  // would NOT compile.  A warning is not good enough.
  //
  // to FPC's credit, while it compiles those incorrect statements, the code
  // it generates for them is correct. (the compiler did what the programmer
  // told it to do.)

  // ---------------------------------------------------------------------------
  // case 3.
  // Delphi emits neither a warning nor an error but, generates correct code.
  // if/when runtime range checking is enabled then the problem will be
  // reported.

  // FPC neither emits a warning nor a hint and generates INCORRECT code.

  // in this case, range checking should be done at runtime and, only if the
  // programmer requested range checking.

  b := 3;
  AnInt := (b * 2) div 3;

  AnArray[AnInt] := SPACE;   // FPC generates INCORRECT code (no sign extension)
                             // effectively turning a signed int into an
                             // unsigned int.   That is INCORRECT (64bit) and
                             // cannot be justified with a range-checking
                             // argument.

                             // Delphi generates correct code.

  // ---------------------------------------------------------------------------
  // case 4
  // same as case 3 but using an int64 instead of an int32

  AnInt64 := (b * 2) div 3;  // for this expression FPC generates correct code
  AnArray[AnInt64] := SPACE; // and doesn't complain about it.

  // ---------------------------------------------------------------------------
  // CONCLUSION:

  // FPC compiles statements that it shouldn't compile (cases 1 and 2)
  // FPC's incorrect handling of sign extension makes
  //
  // AnArray[AnInt] <> AnArray[AnInt64]
  //
  // there is no correct argument that can justify that behavior.


  writeln('press enter/return to end this program');
  readln;
end.
